from unittest import TestCase

from g2p.mappings import Mapping
from g2p.transducer import Transducer

class TransitiveTest(TestCase):
    #TODO: This relies on transitive closure. Refactor to include in g2p from readalongs
    def setUp(self):
        self.test_conversion_data = [
            {'in_lang': 'git',
             'out_lang': 'eng-arpabet',
             'in_text': "K̲'ay",
             'out_text': 'K HH AE Y'},
            {'in_lang': 'git',
             'out_lang': 'eng-arpabet',
             'in_text': "guts'uusgi'y",
             'out_text': 'G UW T S HH UW S G IY HH Y'},
            {'in_lang': 'str-sen',
             'out_lang': 'eng-arpabet',
             'in_text': 'X̱I¸ÁM¸',
             'out_text': 'SH W IY HH EY M HH'},
            {'in_lang': 'ctp',
             'out_lang': 'eng-arpabet',
             'in_text': 'Qneᴬ',
             'out_text': 'HH N EY'}
        ]

    def test_conversions(self):
        ''' Some conversion that were posing problems for readalongs.
            These might fail if the lookup tables change.
        '''
        for test in self.test_conversion_data:
            mapping = Mapping(in_lang=test['in_lang'], out_lang=test['out_lang'])
            transducer = Transducer(mapping)
            conversion = transducer(test['in_text'])
            self.assertEqual(conversion[0], test['out_text'])

    def test_reduced_indices(self):
        mapping = Mapping(in_lang='git', out_lang='eng-arpabet')
        transducer = Transducer(mapping)
        conversion = transducer("K̲'ay")
        self.assertEqual(conversion[1].reduced(), [
                         (2, 2), (3, 5), (4, 8), (5, 9)])         
        conversion1 = transducer("yukwhl")
        self.assertEqual(conversion1[1].reduced(), [
                         (1, 2), (2, 5), (3, 7), (4, 9), (6, 10)])


# -*- coding: utf-8 -*-

import os
import json
from g2p.tests.private.git_data_wrangler import returnLinesFromDocuments
from g2p.tests.private import __file__ as private_dir_f
from g2p.mappings import Mapping
from g2p.transducer import Transducer
from typing import List, Union



class Stats:
    def __init__(self, base_string: str, compare_string: str): 
        self.base_string = base_string
        self.compare_string = compare_string 
    
    def 


class Story:
    '''
    Initialize class with path (json,data,word document).
    Accesses story data.
    '''
    def __init__(self, path: str):
        '''
        The path should be to one of the three:
        - a word document containing the data
        - a json file containing the data
        - add actual data, not sure of file type?
        '''
        self.path = path
        
        if self.path.lower().endswith('.docx'):
            self.data = self.parse_word(self.path)
       
        elif self.path.lower().endswith('.json'):
            self.data = self.parse_json(self.path)
        
        else:
            raise TypeError("Not a supported file type")

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    # TODO: Document or delete!
    def length_of_line(self, line_index, line_key):
        return len(self.data[line_index][line_key])

    def parse_word(self, path: str):
        '''
        Takes the path to a word document.
        Returns returnLinesFromDocuments.
        '''
        return returnLinesFromDocuments([self.path])

    def parse_json(self, path: str):
        '''
        Takes the path to a JSON file.
        Returns the lines as strings.
        '''
        with open(self.path, 'r') as f:
            json_data = json.load(f)
        return json_data

if __name__ == '__main__':
    mapping = Mapping(language={"lang": "git", "table": "Orthography (Deterministic)"}, case_sensitive=False)
    transducer = Transducer(mapping)
    private_dir = os.path.dirname(private_dir_f)
    story_json = Story(os.path.join(private_dir, 'BS - Dihlxw', 'Dihlxw Story 2013-04-29 for HD copy - clean.json'))  
    story_docx = Story(os.path.join(private_dir, 'BS - Dihlxw', 'Dihlxw Story 2013-04-29 for HD copy - clean.docx')) 
    breakpoint()


# Compare characters


# RETURN
# Failure percentage as an integer
# Count failures of the same type



"""
Test all Mappings
"""

from unittest import TestCase

import g2p.create_ipa_mapping as mapper

def make_map_dict(mapping):
    return dict((m['in'], m['out']) for m in mapping['map'])

class TestMappings(TestCase):
    def setUp(self):
        self.target_mapping = {
            "type": "mapping",
            "in_metadata": {
                "lang": "eng-ipa",
                "format": "ipa",
                "delimiter": ""
            },
            "out_metadata": {
                "lang": "eng-arpabet",
                "format": "arpabet",
                "delimiter": " "
            },
            "map": [
                {"in": "ɑ",	"out": "AA"},
                {"in": "eː",	"out": "EY"},
                {"in": "i",	"out": "IY"},
                {"in": "u",	"out": "UW"},
                {"in": "tʃ",	"out": "CH"},
                {"in": "p",	"out": "P"},
                {"in": "t",	"out": "T"},
                {"in": "k",	"out": "K"},
                {"in": "w",	"out": "W"},
            ]
        }

    def test_unigram_mappings(self):
        src_mapping = {
            "type": "mapping",
            "in_metadata": {
                "lang": "crj",
                "format": "custom",
                "delimiter": ""
            },
            "out_metadata": {
                "lang": "crj-ipa",
                "format": "ipa",
                "delimiter": ""
            },
            "map": [
                {"in": "ᐃ", "out": "i"},
                {"in": "ᐅ", "out": "u"},
                {"in": "ᐊ", "out": "a"},
            ],
        }
        mapping = make_map_dict(mapper.create_mapping(src_mapping,
                                                      self.target_mapping))
        self.assertEqual(mapping['a'], 'ɑ')
        self.assertEqual(mapping['i'], 'i')
        self.assertEqual(mapping['u'], 'u')

    def test_bigram_mappings(self):
        src_mapping = {
            "type": "mapping",
            "in_metadata": {
                "lang": "crj",
                "format": "custom",
                "delimiter": ""
            },
            "out_metadata": {
                "lang": "crj-ipa",
                "format": "ipa",
                "delimiter": ""
            },
            "map": [
                {"in": "ᐱ", "out": "pi"},
                {"in": "ᑎ", "out": "ti"},
                {"in": "ᑭ", "out": "ki"},
            ],
        }
        mapping = make_map_dict(mapper.create_mapping(src_mapping,
                                                      self.target_mapping))
        self.assertEqual(mapping['pi'], 'pi')
        self.assertEqual(mapping['ti'], 'ti')
        self.assertEqual(mapping['ki'], 'ki')

    def test_trigram_mappings(self):
        src_mapping = {
            "type": "mapping",
            "in_metadata": {
                "lang": "crj",
                "format": "custom",
                "delimiter": ""
            },
            "out_metadata": {
                "lang": "crj-ipa",
                "format": "ipa",
                "delimiter": ""
            },
            "map": [
                {"in": "ᒋ", "out": "t͡ʃi"},
                {"in": "ᒍ", "out": "t͡ʃu"},
                {"in": "ᒐ", "out": "t͡ʃa"},
            ],
        }
        mapping = make_map_dict(mapper.create_mapping(src_mapping,
                                                      self.target_mapping))
        self.assertEqual(mapping['t͡ʃi'], 'tʃi')
        self.assertEqual(mapping['t͡ʃu'], 'tʃu')
        self.assertEqual(mapping['t͡ʃa'], 'tʃɑ')

    def test_long_mappings(self):
        src_mapping = {
            "type": "mapping",
            "in_metadata": {
                "lang": "crj",
                "format": "custom",
                "delimiter": ""
            },
            "out_metadata": {
                "lang": "crj-ipa",
                "format": "ipa",
                "delimiter": ""
            },
            "map": [
                {"in": "ᐧᐯ", "out": "pʷeː"},
                {"in": "ᐧᑌ", "out": "tʷeː"},
                {"in": "ᐧᑫ", "out": "kʷeː"},
            ],
        }
        mapping = make_map_dict(mapper.create_mapping(src_mapping,
                                                      self.target_mapping))
        self.assertEqual(mapping['pʷeː'], 'pweː')
        self.assertEqual(mapping['tʷeː'], 'tweː')
        self.assertEqual(mapping['kʷeː'], 'kweː')



from typing import Dict, List, Pattern, Tuple, Union

def return_default_mapping(input_strings: List[str], output_strings: List[str],
                            input_index_offsets: List[int], output_index_offsets: List[int]):
    ''' This function takes an arbitrary number of input & output strings and their corresponding index offsets.
        It then zips them up 1 by 1. If the input is longer than the output or vice versa, it continues zipping
        using the last item of either input or output respectively.
    '''
    new_input = {}
    # go through each input or output whichever is longer
    for i in range(0, max(len(input_strings), len(output_strings))):
        try:
            input_i = input_index_offsets[i]
        except IndexError:
            input_i = input_index_offsets[-1]
        try:
            output_i = output_index_offsets[i]
        except IndexError:
            output_i = output_index_offsets[-1]
        try:
            # if inputs and outputs are the same length, just zip them up
            new_input[input_i] = {'input_string': input_strings[i],
                                    'output': {output_i: output_strings[i]}}
        except IndexError:
            # but if the input is longer than output, use the last output character
            if len(input_strings) > len(output_strings):
                new_input[input_i] = {'input_string': input_strings[i],
                                        'output': {output_i: output_strings[-1]}}
            # conversely if the output is longer than input, use the last input character
            elif len(input_strings) < len(output_strings):
                if input_i in new_input:
                    intermediate_output = new_input[input_i]['output']
                else:
                    intermediate_output = {}
                new_input[input_i] = {'input_string': input_strings[-1],
                                        'output': {**intermediate_output, **{output_i: output_strings[i]}}}
    return new_input

def convert_index_to_tuples(index):
    container = []
    for input_index, val in index.items():
        input_string = val['input_string']
        for output_index, output_string in val['output'].items():
            container.append(((input_index, input_string),
                                (output_index, output_string)))
    return container

def convert_tuples_to_index(tuples, reverse=False):
    indices = {}
    for tup in tuples:
        if reverse:
            inp = tup[1]
            outp = tup[0]
        else:
            inp = tup[0]
            outp = tup[1]
        if inp[0] in indices:
            intermediate_output = indices[inp[0]].get('output', {})
        else:
            intermediate_output = {}
        indices[inp[0]] = {'input_string': inp[1],
                            'output': {**intermediate_output, **{outp[0]: outp[1]}}}
    return indices


# -*- coding: utf-8 -*-

from unittest import main, TestCase
import os
from g2p import make_g2p
from g2p.mappings import Mapping
from g2p.transducer import Transducer


class LangTest(TestCase):
    ''' Basic Test for individual lookup tables
    '''

    def setUp(self):
        git = [
            {"in_lang": "git",
                "out_lang": "git-ipa",
                "samples": [
                    ('gwila', '\u0261ʷilæ'),
                    ("hlik\u0332'sxw", "ɬiq\u0294sx\u02b7")
                ],
             },
            {'in_lang': 'git',
                'out_lang': 'eng-arpabet',
                "samples": [
                    ("K̲'ay", 'K HH AE Y'),
                    ("guts'uusgi'y", 'G UW T S HH UW S G IY HH Y')
                ]},
            {'in_lang': 'str-sen',
                'out_lang': 'eng-arpabet',
                "samples": [
                    ('X̱I¸ÁM¸', 'SH W IY HH EY M HH')
                ]},
            {'in_lang': 'ctp',
                'out_lang': 'eng-arpabet',
                "samples": [
                    ('Qneᴬ', 'HH N EY')
                ]}
        ]

        self.langs_to_test = git

    def test_io(self):
        # go through each language declared in the test case set up
        for lang in self.langs_to_test:
            in_lang = lang['in_lang']
            out_lang = lang['out_lang']
            transducer = make_g2p(in_lang, out_lang)
            # go through each table in the current lang
            for sample in lang['samples']:
                # assert that the transduced first item in the tuple is equal to the second item in the tuple
                self.assertEqual(transducer(sample[0]), sample[1])


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
from unittest import main, TestCase
from g2p.mappings import Mapping
from g2p.transducer import CompositeTransducer, Transducer
# Will cause errors on machines without private data
from g2p.tests.private.git_data_wrangler import returnLinesFromDocuments
from g2p.tests.private import __file__ as private_dir
from typing import List
from g2p.log import TEST_LOGGER as logger


def scrub_text(txt: str, to_scrub: List[str] = ['=', '-', '~']) -> str:
    ''' Given some text (txt), scrub all characters in list (to_scrub) from text.
    '''
    for char in to_scrub:
        txt = txt.replace(char, '')
    return txt


class GitTest(TestCase):
    ''' Test for Orthography (Deterministic) to APA
    '''
    @classmethod
    def setUpClass(cls):
        # set up log counter
        cls.formatted_data = returnLinesFromDocuments(
            [
                os.path.join(os.path.dirname(private_dir),
                             'VG - Nass Volcano',
                             '2016-07-12 Nass River Volcano - CF edit VG check2.docx'),
                #os.path.join(os.path.dirname(private_dir),
                #             'VG - Kitwancool Surveyed',
                #             '2014-01-29 VG Kitwancool Reserve Surveyed CAF edit3 for HD.docx'),
                #os.path.join(os.path.dirname(private_dir),
                #             'VG - Founding of Gitanyow',
                #             '2017-01-27 VG The Founding of Git-anyaaw - HD comments more fixes.docx'),
                #os.path.join(os.path.dirname(private_dir),
                #             "HH - Mystery Story",
                #             "15-10-31 HH Hector's Betl'a Betl' - 4 checks_AP.docx"),
                #os.path.join(os.path.dirname(private_dir),
                #             "HH - Win Bekwhl",
                #             "2014-10-22 HH Win Bexwhl K'amk'siwaa - 2016-04-22 CAF 1st gloss.docx"),
                #os.path.join(os.path.dirname(private_dir),
                #             "BS - Dihlxw",
                #             "Dihlxw Story 2013-04-29 for HD copy - clean.docx"),
            ])

        # Declare all of our mappings needed
        cls.orth_to_ipa = Mapping(in_lang='git', out_lang='git-ipa', as_is=True)

        cls.orth_to_ipa_transducer = Transducer(cls.orth_to_ipa)

        cls.ipa_to_orth = Mapping(in_lang='git', out_lang='git-apa', as_is=True, reverse=True)

        cls.ipa_to_orth_transducer = Transducer(cls.ipa_to_orth)

        cls.apa_to_ipa = Mapping(in_lang='git', out_lang='git-apa', as_is=True)

        cls.apa_to_ipa_transducer = Transducer(cls.apa_to_ipa)

        cls.ipa_to_apa = Mapping(in_lang='git', out_lang='git-apa', as_is=True, reverse=True)

        cls.ipa_to_apa_transducer = Transducer(cls.ipa_to_apa)

        cls.orth_to_apa_transducer = CompositeTransducer(
            [cls.orth_to_ipa_transducer, cls.ipa_to_apa_transducer])
        cls.apa_to_orth_transducer = CompositeTransducer(
            [cls.apa_to_ipa_transducer, cls.ipa_to_orth_transducer])

    def test_orth_to_apa(self):     
        for title, story in self.formatted_data.items():
            for line in story:
                try:
                    self.assertEqual(self.orth_to_apa_transducer(
                        line['ortho']), scrub_text(line['apa']))
                except:
                    logger.exception(f"{self.orth_to_apa_transducer(line['ortho'])} is not equal to {scrub_text(line['apa'])}")
        logger.info("The logger came across %s exceptions", logger.exception.count)


    def test_apa_to_orth(self):
        for title, story in self.formatted_data.items():
            for line in story:
                try:
                    self.assertEqual(self.apa_to_orth_transducer(
                        scrub_text(line['apa'])), line['ortho'])
                except:
                    logger.exception(f"{self.orth_to_apa_transducer(line['ortho'])} is not equal to {scrub_text(line['apa'])}")
        logger.info("The logger came across %s exceptions", logger.exception.count)

if __name__ == "__main__":
    main()



'''
Class for performing transductions based on mappings

'''
import re
import copy
from typing import List, Tuple, Union
from collections.abc import Iterable

from g2p.transducer.utils import convert_index_to_tuples, convert_tuples_to_index

# Avoid TypeError in Python < 3.7 (see
# https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object)
copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r

class Indices():
    ''' Class containing input and output states along of a Transducer along with indices
    '''

    def __init__(self, index: Union[dict, List[Tuple[Tuple[int, str], Tuple[int, str]]]]):
        if isinstance(index, dict):
            self.indices = convert_index_to_tuples(index)
        else:
            self.indices = index
        self._input_states = sorted(
            [io[0] for io in self.indices], key=lambda x: x[0])
        self._output_states = sorted(
            [io[1] for io in self.indices], key=lambda x: x[0])
        self.condensed_input_states = [
            list(v) for v in dict(self._input_states).items()]
        self.condensed_output_states = [
            list(v) for v in dict(self._output_states).items()]
        self.filtered_condensed_input_states = [
            list(v) for v in dict(self._input_states).items() if v[1]]
        self.filtered_condensed_output_states = [
            list(v) for v in dict(self._output_states).items() if v[1]]

    def __repr__(self):
        return f"{self.__class__} object with input '{self.input()}' and output '{self.output()}'"

    def __call__(self):
        return self.indices

    def __add__(self, other):
        return Indices(self.indices + other.indices)

    def __iadd__(self, other):
        return Indices(self.indices + other.indices)

    def __iter__(self):
        return iter(self.indices)

    def reduced(self):
        ''' Find how many indices it takes before input and output both move forward by one phone
        '''
        filtered = self.filter_empty_values()
        reduced = []
        intermediate_index = {
            'in': filtered[0][0], 'out': filtered[0][1]}
        for io in filtered:
            inp = io[0]
            outp = io[1]
            if inp == intermediate_index['in'] and outp > intermediate_index['out']:
                intermediate_index['out'] = outp
            if inp > intermediate_index['in'] and outp == intermediate_index['out']:
                intermediate_index['in'] = inp
            if inp > intermediate_index['in'] and outp > intermediate_index['out']:
                intermediate_index = {'in': inp, 'out': outp}
                reduced.append(copy.deepcopy(intermediate_index))
        reduced.append({"in": len(self.filtered_condensed_input_states),
                        "out": len(self.filtered_condensed_output_states)})
        reduced = [(x['in'], x['out']) for x in reduced]
        return reduced

    def filter_empty_values(self):
        filtered = []
        input_offset = 0
        output_offset = 0
        indices = self.indices
        for io in indices:
            if not io[0][1]:
                # epenthesis offsets the input by -1
                input_offset -= 1
            elif not io[1][1]:
                 # deletion offsets the output by -1
                output_offset -= 1
            else:
                filtered.append((io[0][0] + input_offset, io[1][0] + output_offset))
        return filtered

    def input(self):
        """ Return the input of a given transduction
        """
        return ''.join([state[1] for state in self.condensed_input_states])

    def output(self):
        """ Return the output of a given transduction
        """
        return ''.join([state[1] for state in self.condensed_output_states])


class IndexSequence():
    '''Class containing a sequence of IO States
    '''

    def __init__(self, *args: Indices):
        self.states = []
        for arg in args:
            if isinstance(arg, Indices):
                self.states.append(arg)
            if isinstance(arg, IndexSequence):
                self.states += self.unpack_states(arg)

    def __iter__(self):
        return iter(self.states)

    def __repr__(self):
        return f"{self.__class__} object with input '{self.input()}' and output '{self.output()}'"

    def unpack_states(self, seq):
        states = []
        for state in seq.states:
            if isinstance(state, Indices):
                states.append(state)
            if isinstance(state, Iterable) and any([isinstance(x, Indices) for x in state]):
                states += self.unpack_states(state)
        return states

    def compose_filtered_and_reduced_indices(self, i1, i2):
        if not i1:
            return i2
        i2_dict = dict(i2)
        i2_idx = 0
        results = []
        for i1_in, i1_out in i1:
            highest_i2_found = 0 if not results else results[-1][1]
            while i2_idx <= i1_out:
                if i2_idx in i2_dict and i2_dict[i2_idx] > highest_i2_found:
                    highest_i2_found = i2_dict[i2_idx]
                i2_idx += 1
            if results:
                assert(i1_in >= results[-1][0])
                assert(highest_i2_found >= results[-1][1])
            results.append((i1_in, highest_i2_found))
        return results

    def input(self):
        return self.states[0].input()

    def output(self):
        return self.states[-1].output()

    def reduced(self):
        ''' This is a reduced tuple-based format for indices. It requires that Indices in the
        IndexSequence be composed, and then they can be reduced to a list of tuples where each
        tuple contains an input and output index corresponding with that character. The list is equal
        to the length of unique characters
        '''
        composed_states = self.compose_filtered_and_reduced_indices(self.states[0].reduced(), self.states[1].reduced())
        counter = 2
        while counter < len(self.states):
            composed_states = self.compose_filtered_and_reduced_indices(
                composed_states, self.states[counter].reduced())
            counter += 1
        return composed_states



"""
    Unittests for index preservation

"""

from unittest import main, TestCase
from g2p.mappings import Mapping
from g2p.transducer import Transducer
from g2p.transducer.indices import Indices


class IndicesTest(TestCase):
    ''' Basic Transducer Test
    Preserve character-level mappings:

    Test Case #1
        # Simple conversion

        0 1 2 3
        t e s t

        p e s t
        0 1 2 3

        [ ((0, 't'), (0, 'p')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #2:
        # Allow for deletion of segments

        0 1 2 3
        t e s t

        t s t
        0 1 2

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, '')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #3
        # Allow for one-to-many

        0 1 2 3
        t e s t

        c h e s t
        0 1 2 3 4

        [ ((0, 't'), (0, 'c')),
          ((0, 't'), (1, 'h')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (3, 's')),
          ((3, 't'), (4, 't')) ]

    Test Case #4
        # Allow for many-to-one

        0 1 2 3
        t e s t

        p s t
        0 1 2

        [ ((0, 't'), (0, 'p')),
          ((1, 'e'), (0, 'p')),
          ((2, 's'), (1, 's')),
          ((3, 't'), (2, 't')) ]

    Test Case #5
        # Allow for epenthesis
         0 1 2 3
         t e s t

         t e s t y
         0 1 2 3 4

        [ ((-1, 'y'), (4, 'y')),
          ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

     Test Case #6
        # Allow metathesis
         0 1 2 3
         t e s t

         t s e t
         0 1 2 3

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (1, 's')),
          ((3, 't'), (3, 't')) ]

    Test Case #7
        # Allow order-sensitive operations
        0 1 2 3
        t e s t

        t e s h t
        0 1 2 3 4

        t e s t
        0 1 2 3

        AS IS

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((3, 't'), (3, 't')) ]

          or not

        [ ((0, 't'), (0, 't')),
          ((1, 'e'), (1, 'e')),
          ((2, 's'), (2, 's')),
          ((2, 's'), (3, 'h')),
          ((3, 't'), (4, 't')) ]

    Test Case #8
        # Allow multiple processes which alter the indices
        0 1 2 3
        t e s t

        c h e s t
        0 1 2 3 4

        c h e s s
        0 1 2 3 4

        [ ((0, 't'), (0, 'c')),
          ((1, 'e'), (1, 'h')),
          ((1, 'e'), (2, 'e')),
          ((2, 's'), (3, 's')),
          ((3, 't'), (4, 's')) ]

    '''

    def setUp(self):
        self.test_mapping_one = Mapping(
            [{'in': 't', "out": 'p', 'context_after': 'e'}])
        self.test_mapping_two = Mapping([{"in": 'e', "out": ""}])
        self.test_mapping_three = Mapping(
            [{"in": 't', 'out': 'ch', 'context_after': 'e'}])
        self.test_mapping_four = Mapping([{'in': 'te', 'out': 'p'}])
        self.test_mapping_five = Mapping(
            [{'context_before': 't', 'context_after': '$', 'in': '', 'out': 'y'}])
        self.test_mapping_six = Mapping(
            [{"in": "e{1}s{2}", "out": "s{2}e{1}"}]
        )
        self.test_mapping_seven = Mapping(
            [{"in": "s", "out": "sh"}, {"in": "sh", "out": "s"}]
        )
        self.test_mapping_seven_as_is = Mapping(
            [{"in": "s", "out": "sh"}, {"in": "sh", "out": "s"}], as_is=True
        )
        self.test_mapping_eight = Mapping([{"in": "te", "out": "che"},
                                           {"in": "t", "out": "s"}])
        self.test_mapping_combining = Mapping(
            [{'in': 'k{1}\u0313{2}', 'out': "'{2}k{1}"}])
        self.test_mapping_wacky = Mapping(
            [{"in": "\U0001f600{1}\U0001f603\U0001f604{2}\U0001f604{3}",
              "out": "\U0001f604\U0001f604\U0001f604{2}\U0001f604{3}\U0001f604{1}"}]
        )
        self.test_mapping_wacky_lite = Mapping(
            [{"in": "a{1}bc{2}c{3}",
              "out": "ccc{2}c{3}c{1}"}]
        )
        self.test_mapping_circum = Mapping(
            [{'in': 'a{1}c{2}', 'out': 'c{2}a{1}c{2}'}]
        )
        self.trans_one = Transducer(self.test_mapping_one)
        self.trans_two = Transducer(self.test_mapping_two)
        self.trans_three = Transducer(self.test_mapping_three)
        self.trans_four = Transducer(self.test_mapping_four)
        self.trans_five = Transducer(self.test_mapping_five)
        self.trans_six = Transducer(self.test_mapping_six)
        self.trans_seven = Transducer(self.test_mapping_seven)
        self.trans_seven_as_is = Transducer(self.test_mapping_seven_as_is)
        self.trans_eight = Transducer(self.test_mapping_eight)
        self.trans_combining = Transducer(self.test_mapping_combining)
        self.trans_wacky = Transducer(self.test_mapping_wacky)
        self.trans_wacky_lite = Transducer(self.test_mapping_wacky_lite)
        self.trans_circum = Transducer(self.test_mapping_circum)

    def test_no_indices(self):
        """ Test straightforward conversion without returning indices.
        """
        transducer = self.trans_combining('k\u0313am')
        self.assertEqual(transducer, "'kam")

    def test_combining(self):
        """ Test index preserving combining characters
        """
        transducer = self.trans_combining('k\u0313am', index=True)
        self.assertEqual(transducer[0], "'kam")
        self.assertEqual(transducer[1](), [((0, "k"), (1, 'k')),
                                           ((1, '\u0313'),
                                            (0, "'")),
                                           ((2, 'a'),
                                            (2, 'a')),
                                           ((3, 'm'), (3, 'm'))])

    def test_wacky(self):
        """ Test weird Unicode emoji transformation...
        """
        transducer_no_i = self.trans_wacky(
            '\U0001f600\U0001f603\U0001f604\U0001f604')
        self.assertEqual(
            transducer_no_i, '\U0001f604\U0001f604\U0001f604\U0001f604\U0001f604')
        transducer_lite = self.trans_wacky_lite(
            'abcc', index=True)
        self.assertEqual(
            transducer_lite[0], 'ccccc')
        transducer = self.trans_wacky(
            '\U0001f600\U0001f603\U0001f604\U0001f604', index=True)
        self.assertEqual(
            transducer[0], '\U0001f604\U0001f604\U0001f604\U0001f604\U0001f604')
        self.assertEqual(transducer[1](), [
            ((0, '😀'), (4, '😄')),
            ((1, '😃😄'), (0, '😄😄😄')),
            ((3, '😄'), (3, '😄'))])

    def test_circum(self):
        """ Test circumfixing
        """
        transducer = self.trans_circum('ac', index=True)
        self.assertEqual(transducer[0], 'cac')
        self.assertEqual(transducer[1](), [((0, 'a'), (1, 'a')),
                                           ((1, 'c'), (0, 'c')),
                                           ((1, 'c'), (2, 'c'))])

    def test_case_one(self):
        """ Test case one
        """
        transducer = self.trans_one('test', True)
        self.assertEqual(transducer[0], 'pest')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 'p')),
                                           ((1, 'e'),
                                            (1, 'e')),
                                           ((2, 's'),
                                            (2, 's')),
                                           ((3, 't'), (3, 't'))])
        self.assertEqual(transducer[1].reduced(), [
                         (1, 1), (2, 2), (3, 3), (4, 4)])

    def test_case_two(self):
        transducer = self.trans_two('test', True)
        self.assertEqual(transducer[0], 'tst')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 't')),
                                           ((1, 'e'),
                                            (1, '')),
                                           ((2, 's'),
                                            (2, 's')),
                                           ((3, 't'), (3, 't'))])
        self.assertEqual(transducer[1].reduced(), [(2, 1), (3, 2), (4, 3)])

    def test_case_three(self):
        transducer = self.trans_three('test', True)
        self.assertEqual(transducer[0], 'chest')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 'c')),
                                           ((0, 't'),
                                            (1, 'h')),
                                           ((1, 'e'),
                                            (2, 'e')),
                                           ((2, 's'),
                                            (3, 's')),
                                           ((3, 't'), (4, 't'))])

    def test_case_four(self):
        transducer = self.trans_four('test', True)
        self.assertEqual(transducer[0], 'pst')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 'p')),
                                           ((1, 'e'),
                                            (0, 'p')),
                                           ((2, 's'),
                                            (1, 's')),
                                           ((3, 't'), (2, 't'))])

    def test_case_six(self):
        transducer = self.trans_six('test', True)
        self.assertEqual(transducer[0], 'tset')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 't')),
                                           ((1, 'e'), (2, 'e')),
                                           ((2, 's'), (1, 's')),
                                           ((3, 't'), (3, 't'))])

    def test_case_long_six(self):
        # transducer_no_i = self.trans_six('esesse')
        # self.assertEqual(transducer_no_i, 'sesese')
        transducer = self.trans_six('esesse', True)
        self.assertEqual(transducer[0], 'sesese')

    def test_case_seven(self):
        transducer_as_is = self.trans_seven_as_is('test', True)
        self.assertEqual(transducer_as_is[0], 'test')
        self.assertEqual(transducer_as_is[1](), [((0, 't'), (0, 't')),
                                                 ((1, 'e'), (1, 'e')),
                                                 ((2, 's'), (2, 's')),
                                                 ((3, 't'), (3, 't'))])
        transducer = self.trans_seven('test', True)
        self.assertEqual(transducer[0], 'tesht')

        self.assertEqual(transducer[1](), [((0, 't'), (0, 't')),
                                           ((1, 'e'), (1, 'e')),
                                           ((2, 's'), (2, 's')),
                                           ((2, 's'), (3, 'h')),
                                           ((3, 't'), (4, 't'))])

    def test_case_eight(self):
        transducer = self.trans_eight('test', True)
        self.assertEqual(transducer[0], 'chess')
        self.assertEqual(transducer[1](), [((0, 't'), (0, 'c')),
                                           ((1, 'e'), (1, 'h')),
                                           ((1, 'e'), (2, 'e')),
                                           ((2, 's'), (3, 's')),
                                           ((3, 't'), (4, 's'))])

    def test_conversion(self):
        tuple_format = [
            ((0, 'a'), (0, 'b'))
        ]
        dict_format = {0: {'input_string': 'a', 'output': {0: 'b'}}}
        indices_from_tups = Indices(tuple_format)
        indices_from_dict = Indices(dict_format)
        self.assertEqual(indices_from_dict(), indices_from_tups())


if __name__ == "__main__":
    main()



"""

Module for all things related to lookup tables

"""

import csv
import os
import re
import json
from typing import DefaultDict, List, Pattern, Union
from collections import defaultdict, OrderedDict
from itertools import chain
from pathlib import Path
from operator import methodcaller

import yaml

from g2p import exceptions
from g2p.mappings.langs import __file__ as LANGS_FILE, LANGS, MAPPINGS_AVAILABLE
from g2p.mappings.utils import create_fixed_width_lookbehind, escape_special_characters, normalize
from g2p.mappings.utils import flatten_abbreviations, IndentDumper, load_abbreviations_from_file
from g2p.mappings.utils import load_from_file, load_mapping_from_path, unicode_escape, validate
from g2p.log import LOGGER


class Mapping():
    """ Class for lookup tables

        @param as_is: bool = False
            Evaluate g2p rules in mapping in the order they are.
            Default is to reverse sort them by length.

        @param case_sensitive: bool = True
            Lower all rules and conversion input

        @param escape_special: bool = False
            Escape special characters in rules

        @param norm_form: str = "NFC"
            Normalization standard to follow. NFC | NKFC | NFD | NKFD

        @param reverse: bool = False
            Reverse all mappings

    """

    def __init__(self, mapping=None, abbreviations: Union[str, DefaultDict[str, List[str]]] = False, **kwargs):
        # should these just be explicit instead of kwargs...
        self.allowable_kwargs = ['language_name', 'display_name', 'mapping', 'in_lang',
                                 'out_lang', 'out_delimiter', 'as_is', 'case_sensitive', 'escape_special', 'norm_form', 'reverse']
        self.kwargs = OrderedDict(kwargs)
        self.processed = False
        if isinstance(abbreviations, defaultdict) or not abbreviations:
            self.abbreviations = abbreviations
        elif abbreviations:
            self.abbreviations = load_abbreviations_from_file(
                abbreviations)
        # Handle user-supplied list
        if isinstance(mapping, list):
            self.mapping = validate(mapping)
        elif isinstance(mapping, str):
            loaded_config = load_mapping_from_path(mapping)
            self.process_loaded_config(loaded_config)
        else:
            if "in_lang" in self.kwargs and "out_lang" in self.kwargs:
                loaded_config = self.find_mapping(
                    self.kwargs['in_lang'], self.kwargs['out_lang'])
                self.process_loaded_config(loaded_config)
            elif 'id' in self.kwargs:
                loaded_config = self.find_mapping_by_id(self.kwargs['id'])
                self.process_loaded_config(loaded_config)
            else:
                raise exceptions.MalformedLookup()
        if self.abbreviations:
            for abb, stands_for in self.abbreviations.items():
                abb_match = re.compile(abb)
                abb_repl = '|'.join(stands_for)
                if self.mapping and 'match_pattern' not in self.mapping[0]:
                    for io in self.mapping:
                        for key in io.keys():
                            if key in ['in', 'out', 'context_before', 'context_after'] and re.search(abb_match, io[key]):
                                io[key] = re.sub(
                                    abb_match, unicode_escape(abb_repl), io[key])
        if not self.processed:
            self.mapping = self.process_kwargs(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __call__(self):
        return self.mapping

    def __iter__(self):
        return iter(self.mapping)

    def inventory(self, in_or_out: str = 'in'):
        ''' Return just inputs or outputs as inventory of mapping
        '''
        return [x[in_or_out] for x in self.mapping]

    def process_loaded_config(self, config):
        ''' For a mapping loaded from a file, take the keyword arguments and supply them to the
            Mapping, and get any abbreviations data.
        '''
        self.mapping = config['mapping_data']
        mapping_kwargs = OrderedDict(
            {k: v for k, v in config.items() if k in self.allowable_kwargs})
        self.abbreviations = config.get('abbreviations_data', None)
        # Merge kwargs, but prioritize kwargs that initialized the Mapping
        self.kwargs = {**mapping_kwargs, **self.kwargs}

    def plain_mapping(self):
        ''' Return mapping
        '''
        return [{k: v for k, v in io.items() if k in ['in', 'out', 'context_before', 'context_after']} for io in self.mapping]

    def process_kwargs(self, mapping):
        ''' Apply kwargs in the order they are provided. kwargs are ordered as of python 3.6
        '''
        # Add defaults
        if 'as_is' not in self.kwargs:
            self.kwargs['as_is'] = False
        if 'case_sensitive' not in self.kwargs:
            self.kwargs['case_sensitive'] = True
        if 'escape_special' not in self.kwargs:
            self.kwargs['escape_special'] = False
        if 'norm_form' not in self.kwargs:
            self.kwargs['norm_form'] = 'NFC'
        if 'reverse' not in self.kwargs:
            self.kwargs['reverse'] = False
        # Process kwargs in order received
        for kwarg, val in self.kwargs.items():
            if kwarg == 'as_is' and not val:
                # sort by reverse len
                mapping = sorted(mapping, key=lambda x: len(
                    x["in"]), reverse=True)
            if kwarg == 'escape_special' and val:
                mapping = [escape_special_characters(x) for x in mapping]
            if kwarg == 'case_sensitive' and not val:
                mapping = self.lower_mappings(mapping)
            if kwarg == 'norm_form' and val:
                for io in mapping:
                    for k, v in io.items():
                        if isinstance(v, str):
                            io[k] = normalize(v, self.kwargs['norm_form'])
            if kwarg == 'reverse' and val:
                mapping = self.reverse_mappings(mapping)
        # After all processing is done, turn into regex
        for io in mapping:
            io['match_pattern'] = self.rule_to_regex(io)
        self.processed = True
        return mapping

    def rule_to_regex(self, rule: str) -> Pattern:
        """Turns an input string (and the context) from an input/output pair
        into a regular expression pattern"""
        if "context_before" in rule and rule['context_before']:
            before = rule["context_before"]
        else:
            before = ''
        if 'context_after' in rule and rule['context_after']:
            after = rule["context_after"]
        else:
            after = ''
        input_match = re.sub(re.compile(
            r'{\d+}'), "", rule['in'])
        try:
            inp = create_fixed_width_lookbehind(before) + input_match
            if after:
                inp += f"(?={after})"
            if not self.kwargs['case_sensitive']:
                rule_regex = re.compile(inp, re.I)
            else:
                rule_regex = re.compile(inp)
        except:
            LOGGER.error(f'Your regex in mapping between {self.kwargs["in_lang"]} and {self.kwargs["out_lang"]} is malformed. \
                    Do you have un-escaped regex characters in your input {inp}, contexts {before}, {after}?')
            raise Exception(
                f'Your regex in mapping between {self.kwargs["in_lang"]} and {self.kwargs["out_lang"]} is malformed. \
                    Do you have un-escaped regex characters in your input {inp}, contexts {before}, {after}?')
        return rule_regex

    def reverse_mappings(self, mapping):
        ''' Reverse the mapping
        '''
        for io in mapping:
            io['in'], io['out'] = io['out'], io['in']
        return mapping

    def lower_mappings(self, mapping):
        ''' Lower the mapping
        '''
        for io in mapping:
            for k, v in io.items():
                if k in ['in', 'out', 'context_before', 'context_after']:
                    io[k] = v.lower()
        return mapping

    def add_abbreviations(self, abbs, mappings):
        ''' Return abbreviated forms, given a list of abbreviations.

        {'in': 'a', 'out': 'b', 'context_before': 'V', 'context_after': '' }
        {'abbreviation': 'V', 'stands_for': ['a','b','c']}
        ->
        {'in': 'a', 'out': 'b', 'context_before': 'a|b|c', 'context_after': ''}
        '''
        for abb in abbs:
            for io in mappings:
                for key in io.keys():
                    if io[key] == abb['abbreviation']:
                        io[key] = abb['stands_for']
        return mappings

    def find_mapping(self, in_lang: str, out_lang: str) -> list:
        ''' Given an input and output, find a mapping to get between them.
        '''
        for mapping in MAPPINGS_AVAILABLE:
            map_in_lang = mapping.get('in_lang', '')
            map_out_lang = mapping.get('out_lang', '')
            if map_in_lang == in_lang and map_out_lang == out_lang:
                return mapping
        raise exceptions.MappingMissing(in_lang, out_lang)

    def find_mapping_by_id(self, map_id: str):
        ''' Find the mapping with a given ID
        '''
        for mapping in MAPPINGS_AVAILABLE:
            if mapping.get('id', '') == map_id:
                return mapping

    def mapping_to_file(self, output_path: str, file_type: str):
        ''' Write mapping to file
        '''
        if not os.path.isdir(output_path):
            raise Exception("Path %s is not a directory", output_path)
        fn = os.path.join(output_path, self.kwargs.get('in_lang', 'und') + "_to_" +
                          self.kwargs.get('out_lang', 'und') + "." + file_type)
        fieldnames = ['in', 'out', 'context_before', 'context_after']
        filtered = [{k: v for k, v in io.items() if k in fieldnames}
                    for io in self.mapping]
        if file_type == 'json':
            with open(fn, 'w') as f:
                json.dump(filtered, f, indent=4)
        elif file_type == 'csv':
            with open(fn, 'w') as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                for io in filtered:
                    writer.writerow(io)
        else:
            raise exceptions.IncorrectFileType

    def config_to_file(self, output_path: str, mapping_type: str):
        ''' Write config to file
        '''
        if not os.path.isdir(output_path):
            raise Exception("Path %s is not a directory", output_path)
        fn = os.path.join(output_path, 'config.yaml')
        template = {"mappings": [
            {
                "language_name": self.kwargs.get('language_name', self.kwargs.get('in_lang', 'und')),
                "display_name": self.kwargs.get('display_name', self.kwargs.get('in_lang', 'und') + " to " + self.kwargs.get('out_lang', 'und')),
                "in_lang": self.kwargs.get('in_lang', 'und'),
                "out_lang": self.kwargs.get('out_lang', 'und'),
                "authors": self.kwargs.get('authors', ['generated']),
                "as_is": self.kwargs.get('as_is', False),
                "case_sensitive": self.kwargs.get('case_sensitive', True),
                "escape_special": self.kwargs.get('escape_special', False),
                "norm_form": self.kwargs.get('norm_form', "NFC"),
                "reverse": self.kwargs.get('reverse', False),
                "mapping": self.kwargs.get('in_lang', 'und') + "_to_" +
                self.kwargs.get('out_lang', 'und') + "." + mapping_type
            }
        ]}
        with open(fn, 'w') as f:
            yaml.dump(template, f, Dumper=IndentDumper,
                      default_flow_style=False)

    def to_file(self, output_path: str, mapping_type: str = 'csv'):
        self.mapping_to_file(output_path, mapping_type)
        self.config_to_file(output_path, mapping_type)


if __name__ == '__main__':
    pass




# -*- coding: utf-8 -*-
# to import
import regex as re
import os
import json
from typing import Dict, List
from docx import Document
from g2p.mappings import Mapping
from g2p.transducer import Transducer


def normalize_string(to_norm: str) -> str:
    ''' Strip the text and normalize it
    '''
    mappings = Mapping(in_lang='git', out_lang='git-ipa')
    transducer = Transducer(mappings)
    norm = transducer(to_norm.strip())
    return norm


def getText(doc: str) -> List[str]:
    ''' Parse text from docx into a list of lines.
        :param doc: Docx file to read
        :returns: List of lines with type List[str] 
    '''
    language_data = []
    doc = Document(doc)
    for para in doc.paragraphs:
      # only add text that isn't empty
        if para.text and not (re.match(r'\d\.', para.text) and len(para.text) < 3):
            language_data.append(str(para.text))  # can't be used on strings
    return language_data


def returnLines(language_data: List[str]) -> List[Dict[str, str]]:
    ''' Turn lines into a list of dicts that represent
        the interlinear nature of the documents ortho.
        :param language_data: List of strings from docx file
        :returns: List of dictionaries containing:
            - ortho (orthographic transcription)
            - morph (morphological breakdown)
            - apa (APA)
            - gloss (gloss)
            - eng (English translation)
    '''
    LINES = []
    if language_data:
        LINE_DICT = {}
        index = 0
        for line in language_data[1:]:
            line = normalize_string(line)
            # If we are at the beginning of a line.
            beginning = len(LINE_DICT.keys()) == 0
            # Must match FULL line, can contain any lowercase/uppercase or combining lowline \u0332
            # Must not start with an apostrophe UNLESS it is for a resonant
            ortho_text = re.search(
                r'^([^\']|(\'N|\'M|\'L|\'W))[a-zA-Z\,\.\:\'\s\u0332\-\–\…\"\?]+[^\']$', line)
            # Should match orthography line if at beginning
            if beginning and not ortho_text:
                print(f'Warning! Did not match Orthography in first line of line {index}')
            elif ortho_text:
                # If we have already read in a line, append to the lines list and continue
                if 'ortho' in LINE_DICT and len(LINE_DICT.keys()) == 5:
                    LINES.append(LINE_DICT)
                    LINE_DICT = {}
                if 'ortho' not in LINE_DICT:
                    LINE_DICT['ortho'] = line
                    continue

            # Gloss must have a number or upper case character in contrast to morph line
            gloss_text = re.search(
                r'(\=cn|cn\=|[1-3](sg|pl)?\.i{1,3}|3sg)', line
                # r'^[a-zA-Z0-9\,\.\'\=\~\-\s]*[^\']*(cn|iii)[a-zA-Z0-9\,\.\'\=\~\-\[\]\s]*([^\']|(tl\'|t\'|k\'|k\u0332\'|ts\'))*$', line
                # r'^[^\'][a-zA-Z0-9\,\.\'\=\~\-\s]+[^\']+([1-4]|cn)[a-zA-Z0-9\,\.\'\=\~\-\[\]\s]+([^\']|(tl\'|t\'|k\'|k\u0332\'|ts\'))$', line
            )
            if gloss_text:
                if 'gloss' in LINE_DICT and LINE_DICT['gloss']:
                    LINE_DICT['gloss'] += ' ' + line
                else:
                    LINE_DICT['gloss'] = line
                continue

            # If there's a schwa, it must be APA line
            apa_text = re.search(
                r'\u0259|\u0294|\u0313|\u0259|\u026c|\u02b7|\u03A7', line)
            if apa_text:
                if 'apa' in LINE_DICT and LINE_DICT['apa']:
                    LINE_DICT['apa'] += ' ' + line
                else:
                    LINE_DICT['apa'] = line
                continue

            # Similar to orthography regex, but lower case.
            morph_text = re.search(
                r'^([^\']|(\'n|\'m|\'l|\'w))[a-zA-Z\,\.\'\=\~\-\?\s\u0332]+([^\']|(tl\'|t\'|k\'|k\u0332\'|ts\'))$', line)
            if morph_text:
                if "morph" in LINE_DICT and LINE_DICT['morph']:
                    LINE_DICT['morph'] += ' ' + line
                else:
                    LINE_DICT['morph'] = line
                continue

            # English line is contained in single quotes at the beginning and end of line
            eng_text = re.search(
                r"^\'[a-zA-Z0-9\-\,\'\?\!\"\./\:\;\s\(\)\…\u2014\u0332\u2013]+\'$", line)
            if eng_text:
                if 'eng' in LINE_DICT and LINE_DICT['eng']:
                    LINE_DICT['eng'] += ' ' + line
                else:
                    LINE_DICT['eng'] = line
                continue
            index += 1
    LINES.append(LINE_DICT)
    return LINES


def returnLinesFromDocuments(docs: List[str]) -> Dict[str, List[Dict[str, str]]]:
    ''' Given a list of paths to docx documents, read the lines from them and parse them
        into the interlinear formatted Dict format.
        :param docs: A list of paths to docx files
    '''
    doc_dict = {}
    for doc in docs:
        data = getText(doc)
        lines = returnLines(data)
        fn, ext = os.path.splitext(doc)
        doc_dict[fn] = lines
    return doc_dict


if __name__ == '__main__':
    formatted = returnLinesFromDocuments(
        [
            os.path.join(
                'VG - Nass Volcano', '2016-07-12 Nass River Volcano - CF edit VG check2.docx'),
            os.path.join('VG - Kitwancool Surveyed',
                         '2014-01-29 VG Kitwancool Reserve Surveyed CAF edit3 for HD.docx'),
            os.path.join('VG - Founding of Gitanyow',
                         '2017-01-27 VG The Founding of Git-anyaaw - HD comments more fixes.docx'),
            os.path.join("HH - Mystery Story",
                         "15-10-31 HH Hector's Betl'a Betl' - 4 checks_AP.docx"),
            os.path.join(
                "HH - Win Bekwhl", "2014-10-22 HH Win Bexwhl K'amk'siwaa - 2016-04-22 CAF 1st gloss.docx"),
            os.path.join(
                "BS - Dihlxw", "Dihlxw Story 2013-04-29 for HD copy - clean.docx"),
        ])
    for k, v in formatted.items():
        with open(k+'.json', 'w', encoding='utf8') as f:
            json.dump(v, f)
        with open(k+'.txt', 'w', encoding='utf8') as f:
            all_ortho = '\n'.join([l['ortho'] for l in v])
            f.write(all_ortho)
        with open(k+'.xml', 'w', encoding='utf8') as f:
            all_ortho = '\n'.join(['<s>' + l['ortho'] + '</s>' for l in v])
            xml = f"<?xml version='1.0' encoding='utf-8'?><TEI><text xml:lang='git'><body><div type='page'><p>{all_ortho}</p></div></body></text></TEI>"
            f.write(xml)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

########################################################
# © Patrick Littell
# 
# Transpile mapping to FST source
#
# We're not currently using this module, but
# once we have index-preserving finite-state G2P,
# this will take the mapping files and turn them into
# FSTs.
#
# AP TODO: This has to be refactored to work with g2p.mappings.Mapping objects
########################################################

from __future__ import print_function, unicode_literals, division
from io import open
import argparse
import json


def make_header_comments(mapping):
    lines = []
    lines.append("#################################")
    lines.append("#")
    lines.append("# FST converting from %s-%s to %s-%s"
                 % (mapping["in_metadata"]["lang"],
                    mapping["in_metadata"]["orth"],
                    mapping["out_metadata"]["lang"],
                    mapping["out_metadata"]["orth"]))
    lines.append("#")
    authors = " & ".join(mapping["authors"])
    lines.append("# Based on a mapping by %s " % (authors,))
    lines.append("# created %s" % mapping["created"])
    lines.append("# last modified %s" % mapping["last_modified"])
    lines.append("#")
    lines.append("#################################")
    lines.append('')
    return lines


def escape_xfst_specials(s):
    return s.replace(
        "-", "%-").replace(
            "0", "%0").replace(
                "|", "%|").replace(":", "%:")


def camelize(s):
    return s[0].upper() + s[1:].lower()


def get_identifier_prefix(mapping):
    return


def compile_correspondences(mapping):
    lines = []
    corrs = [(escape_xfst_specials(x["in"]), escape_xfst_specials(x["out"]))
             for x in mapping["map"]]
    corrs = ["{%s}:{%s}" % (x, y) for x, y in corrs]
    identifier = camelize(mapping["in_metadata"]["lang"]) + \
        camelize(mapping["in_metadata"]["orth"]) + "To" + \
        camelize(mapping["out_metadata"]["lang"]) + \
        camelize(mapping["out_metadata"]["orth"])
    lines.append("define " + identifier + " [" + "|".join(corrs) + "]* ;")
    lines.append("push " + identifier)
    return lines


def go(mapping_filename, output_filename):
    with open(mapping_filename, "r", encoding="utf-8") as fin:
        mapping = json.load(fin)
    lines = "\n".join(make_header_comments(mapping) +
                      compile_correspondences(mapping))
    with open(output_filename, "w", encoding="utf-8") as fout:
        fout.write(lines)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Create an inventory file from a mapping file')
    parser.add_argument('mapping', type=str, help='Mapping filename')
    parser.add_argument('output', type=str, help='Output inventory filename')
    args = parser.parse_args()
    go(args.mapping, args.output)



"""

Views and config to the g2p Studio web app

"""
import sys
import io

from networkx import shortest_path
from networkx.exception import NetworkXNoPath
from flask import Flask, render_template
from flask_socketio import SocketIO, emit
from flask_talisman import Talisman

from g2p.mappings import Mapping
from g2p.mappings.langs import LANGS, LANGS_NETWORK
from g2p.transducer import CompositeTransducer, Transducer
from g2p.transducer.indices import Indices
from g2p.mappings.utils import expand_abbreviations, flatten_abbreviations
from g2p._version import VERSION
from g2p.log import LOGGER

if sys.stdout.encoding != 'utf8':
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="utf8")

if sys.stderr.encoding != 'utf8':
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="utf8")

APP = Flask(__name__)
SOCKETIO = SocketIO(APP)
DEFAULT_N = 10

def return_echart_data(indices: Indices):
    input_string = indices.input()
    input_x = 300
    input_y = 300

    output_string = indices.output()
    output_x = 500
    output_y = 300

    inputs = [{'name': f"{x} (in-{i})", "x": input_x, "y": input_y + (i*50)} for i,x in enumerate(input_string)]
    outputs = [{'name': f"{x} (out-{i})", "x": output_x, "y": output_y + (i*50)} for i,x in enumerate(output_string)]

    data = inputs + outputs

    links = [{"source": x[0][0], "target": x[1][0] + len(input_string)} for x in indices()]

    return data, links

def return_empty_mappings(n=DEFAULT_N):
    ''' Return 'n' * empty mappings
    '''
    y = 0
    mappings = []
    while y < n:
        mappings.append({
            "in": '',
            "out": '',
            "context_before": '',
            "context_after": ''
        })
        y += 1
    return mappings

def hot_to_mappings(hot_data):
    ''' Parse data from HandsOnTable to Mapping format
    '''
    return [{"context_before": x[2], "in": x[0], "context_after": x[3],
             "out": x[1]} for x in hot_data if x[0] and x[1]]

@APP.route('/')
def home():
    """ Return homepage of g2p Studio
    """
    return render_template('index.html', langs=LANGS)

@SOCKETIO.on('index conversion event', namespace='/convert')
def index_convert(message):
    """ Convert input text and return output with indices for echart
    """
    mappings = Mapping(hot_to_mappings(message['data']['mappings']), abbreviations=flatten_abbreviations(
        message['data']['abbreviations']), **message['data']['kwargs'])
    transducer = Transducer(mappings)
    output_string, indices = transducer(message['data']['input_string'], index=True)
    data, links = return_echart_data(indices)
    emit('index conversion response', {'output_string': output_string, 'index_data': data, 'index_links': links})

@SOCKETIO.on('conversion event', namespace='/convert')
def convert(message):
    """ Convert input text and return output
    """
    mappings = Mapping(hot_to_mappings(message['data']['mappings']), abbreviations=flatten_abbreviations(
        message['data']['abbreviations']), **message['data']['kwargs'])
    transducer = Transducer(mappings)
    output_string = transducer(message['data']['input_string'])
    emit('conversion response', {'output_string': output_string})


@SOCKETIO.on('table event', namespace='/table')
def change_table(message):
    """ Change the lookup table
    """
    if message['in_lang'] == 'custom' or message['out_lang'] == 'custom':
        mappings = Mapping(return_empty_mappings())
    else:
        mappings = Mapping(
            in_lang=message['in_lang'], out_lang=message['out_lang'])
    emit('table response', {'mappings': mappings.plain_mapping(),
                            'abbs': expand_abbreviations(mappings.abbreviations),
                            'kwargs': mappings.kwargs})


@SOCKETIO.on('connect', namespace='/connect')
def test_connect():
    """ Let client know disconnected
    """
    emit('connection response', {'data': 'Connected'})


@SOCKETIO.on('disconnect', namespace='/connect')
def test_disconnect():
    """ Let client know disconnected
    """
    print('client disconnected')


def make_g2p(in_lang: str, out_lang: str):
    # Check in_lang is a node in network
    if not in_lang in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called {in_lang}. Please try again.")
        raise(FileNotFoundError)
    
    # Check out_lang is a node in network
    if not out_lang in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called {out_lang}. Please try again.")
        raise(FileNotFoundError)

    # Try to find the shortest path between the nodes
    try:
        path = shortest_path(LANGS_NETWORK, in_lang, out_lang)
    except NetworkXNoPath:
        LOGGER.error(f"Sorry, we couldn't find a way to convert {in_lang} to {out_lang}. Please update your langs by running `g2p update` and try again.")
        raise(NetworkXNoPath)

    # Find all mappings needed
    mappings_needed = []
    for i, lang in enumerate(path):
        try:
            mapping = Mapping(in_lang=path[i], out_lang=path[i+1])
            LOGGER.info(f"Adding mapping between {path[i]} and {path[i+1]} to composite transducer.")
            mappings_needed.append(mapping)
        except IndexError:
            continue
    
    # Either return Transducer or Composite Transducer
    if len(mappings_needed) == 1:
        return Transducer(mappings_needed[0])
    else:
        return CompositeTransducer([Transducer(x) for x in mappings_needed])

    


""" Test Mapping utility functions
"""

import os
from unittest import main, TestCase
from collections import defaultdict

from g2p.mappings import utils
from g2p.tests.public import PUBLIC_DIR
from g2p.exceptions import IncorrectFileType, MalformedMapping
from g2p.transducer.utils import convert_index_to_tuples, convert_tuples_to_index


class UtilsTest(TestCase):
    def setUp(self):
        pass

    def test_abb_flatten_and_expand(self):
        test_rows = [
            ["VOWEL", 'a', 'e', 'i', 'o', 'u'],
            ["OTHER", 't', 'e', 's', 't']
        ]
        default_dict = defaultdict(list)
        default_dict['VOWEL'].extend(['a', 'e', 'i', 'o', 'u'])
        default_dict['OTHER'].extend(['t', 'e', 's', 't'])
        empty_rows = []
        while len(empty_rows) < 10:
            empty_rows.append(['', '', '', '', '', ''])
        self.assertEqual(utils.flatten_abbreviations(test_rows), default_dict)
        self.assertEqual(utils.expand_abbreviations(default_dict), test_rows)
        self.assertEqual(utils.expand_abbreviations({}), empty_rows)

    def test_unicode_escape(self):
        ''' Should turn \u0331 declared in CSVs
            into actual Unicode string for that codepoint
        '''
        self.assertEqual('\u0000', utils.unicode_escape('\\u0000'))
        self.assertEqual('\u0331', utils.unicode_escape('\\u0331'))
        self.assertEqual('\u26F0', utils.unicode_escape('\\u26F0'))

    def test_fixed_width(self):
        # TODO: Test utils.create_fixed_width_lookbehind and utils.pattern_to_fixed_width_lookbehinds
        pass

    def test_pattern(self):
        pass

    def test_load_mapping(self):
        with self.assertRaises(MalformedMapping):
            utils.load_mapping_from_path(os.path.join(
                PUBLIC_DIR, 'mappings', 'malformed_config.yaml'))
        minimal = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_config.yaml'))
        csv = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 0)
        json = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 1)
        xlsx = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 2)
        self.assertEqual(minimal['mapping_data'], csv['mapping_data'])
        self.assertEqual(minimal['mapping_data'], json['mapping_data'])
        self.assertEqual(minimal['mapping_data'], xlsx['mapping_data'])

    def test_validate(self):
        pass

    def test_escape_special(self):
        self.assertEqual(utils.escape_special_characters({'in': '?'}), {'in': '\?'})

    def test_load_abbs(self):
        with self.assertRaises(IncorrectFileType):
            utils.load_abbreviations_from_file(os.path.join(
                PUBLIC_DIR, 'mappings', 'abbreviations.json'))
        abbs = utils.load_abbreviations_from_file(
            os.path.join(PUBLIC_DIR, 'mappings', 'abbreviations.csv'))
        self.assertTrue("VOWEL" in abbs)
        self.assertEqual(abbs['VOWEL'], ['a', 'e', 'i', 'o', 'u'])

    def test_tuple_dict_conversion(self):
        tuple_format = [
            ((0, 'a'), (0, 'b'))
        ]
        dict_format = {0: {'input_string': 'a', 'output': {0: 'b'}}}
        self.assertEqual(convert_index_to_tuples(dict_format), tuple_format)
        self.assertEqual(convert_tuples_to_index(tuple_format), dict_format)


if __name__ == '__main__':
    main()



'''
Class for performing transductions based on mappings

'''

import re
import copy
from typing import Dict, List, Pattern, Tuple, Union
from collections import OrderedDict
from collections.abc import Iterable
from g2p.mappings import Mapping
from g2p.mappings.utils import create_fixed_width_lookbehind, normalize
from g2p.exceptions import MalformedMapping
from g2p.log import LOGGER
from g2p.transducer.indices import Indices, IndexSequence
from g2p.transducer.utils import return_default_mapping

# Avoid TypeError in Python < 3.7 (see
# https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object)
copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r


class Transducer():
    ''' A class for performing transductions based on mappings


    Attributes
    ----------

    mapping: Mapping
        Formatted input/output pairs using the g2p.mappings.Mapping class

    _index_match_pattern: Pattern
        Pattern to match the digit inside curly brackets { } as is this package's convention

    _char_match_pattern: Pattern
        Pattern to match the character(s) preceding the _index_match_pattern

    '''

    def __init__(self, mapping: Mapping):
        self.mapping = mapping
        self.case_sensitive = mapping.kwargs['case_sensitive']
        self.norm_form = mapping.kwargs.get('norm_form', 'none')
        self.out_delimiter = mapping.kwargs.get('out_delimiter', '')
        self._index_match_pattern = re.compile(r'(?<={)\d+(?=})')
        self._char_match_pattern = re.compile(r'[^0-9\{\}]+(?={\d+})', re.U)

    def __repr__(self):
        return f"{__class__} between {self.mapping.kwargs['in_lang']} and {self.mapping.kwargs['out_lang']}"

    def __call__(self, to_convert: str, index: bool = False, debugger: bool = False):
        return self.apply_rules(to_convert, index, debugger)

    def strings_to_lists(self, input_string: str, output_string: str,
                         input_index: int, output_index: int):
        ''' Create default lists
        '''
        # separate string into chars
        default_inputs = [x for x in input_string]
        default_outputs = [x for x in output_string]
        # get offsets for each char
        default_input_offsets = [
            i + input_index for i, v in enumerate(default_inputs)]
        default_output_offsets = [
            i + output_index for i, v in enumerate(default_outputs)]
        default_index = return_default_mapping(
            default_inputs, default_outputs, default_input_offsets, default_output_offsets)
        return default_index

    def return_index(self, input_index: int, output_index: int,
                     input_string: str, output_string: str, original_str: str,
                     intermediate_index: dict) -> Dict[int, Dict[str, str]]:
        """ Return a dictionary containing input indices as keys.

        @param input_index: int
            This is where the input is currently at in the parent loop

        @param output_index: int
            This is where the output is currently at in the parent loop

        @param input_string: str
            This is the input string to convert (can refelct an intermediate stage)

        @param output: str
            This is the output string to convert to

        @param original_str: str
            This is the original input

        @param intermediate_index: dict
            This is a dict containing the intermediate form of the index


        There are four main cases. Empty strings are still treated as having indices,
        which is why the cases are written as (n)one.
        This deals for index-preserving epenthesis and deletion.

        TODO: potentially refactor this to lean more on the return_default_mapping method
         """
        intermediate_index = copy.deepcopy(intermediate_index)
        # (n)one-to-(n)one
        if len(input_string) <= 1 and len(output_string) <= 1:
            # create output dictionary
            new_output = {}
            new_output[output_index] = output_string
            # attach it to intermediate_index and merge output
            intermediate_output = intermediate_index[input_index].get('output', {
            })
            intermediate_index[input_index]['output'] = {**intermediate_output,
                                                         **new_output}
            return intermediate_index

        # (n)one-to-many
        if len(input_string) <= 1 and len(output_string) > 1:
            new_output = {}
            intermediate_output = intermediate_index[input_index].get(
                'output', {})
            for index, output_char in enumerate(output_string):
                new_output[output_index + index] = output_char

            # attach it to intermediate_index and merge output
            if new_output:
                intermediate_index[input_index]['output'] = {**intermediate_output,
                                                             **new_output}
            return intermediate_index

        # many-to-(n)one
        if len(input_string) > 1 and len(output_string) <= 1:
            new_input = {}
            new_output = {output_index: output_string}
            # TODO: do we need intermediate output?
            for index, input_char in enumerate(input_string):
                # prevent feeding rules from leaving traces
                if original_str[index + input_index] == input_char:
                    new_input[input_index + index] = {'input_string': input_char,
                                                      'output': new_output}

            return {**intermediate_index, **new_input}

        # many-to-many
        if len(input_string) > 1 and len(output_string) > 1:
            new_input = {}
            # If indices are explicitly listed with {} notation
            if any(self._char_match_pattern.finditer(input_string)) and any(self._char_match_pattern.finditer(output_string)):
                input_char_matches = [x.group()
                                      for x in self._char_match_pattern.finditer(input_string)]

                input_match_indices = [
                    x.group() for x in self._index_match_pattern.finditer(input_string)]
                inputs = [{'match_index': m, 'string': input_char_matches[i]}
                          for i, m in enumerate(input_match_indices)]
                output_char_matches = [x.group()
                                       for x in self._char_match_pattern.finditer(output_string)]
                output_match_indices = [
                    x.group() for x in self._index_match_pattern.finditer(output_string)]
                outputs = [{'match_index': m, 'string': output_char_matches[i]}
                           for i, m in enumerate(output_match_indices)]
                for match_index in input_match_indices:
                    # Get strings from inputs if they match the match_index
                    explicit_inputs = [x['string']
                                       for x in inputs if x['match_index'] == match_index]
                    # Get strings from outputs if they match the match_index
                    explicit_outputs = [x['string']
                                        for x in outputs if x['match_index'] == match_index]
                    # Get offset for inputs by adding the length of the input string up to the match
                    # plus the overall input index/offset
                    explicit_input_offsets = [
                        len(''.join([x['string'] for x in inputs[:i]])) + input_index for i, v in enumerate(inputs) if v['match_index'] == match_index]
                    # Get offset for outputs by adding the length of the output string up to the match
                    # plus the overall output index/offset
                    explicit_output_offsets = [
                        len(''.join([x['string'] for x in outputs[:i]])) + output_index for i, v in enumerate(outputs) if v['match_index'] == match_index]
                    # Use default mapping to zip them up
                    explicit_index = return_default_mapping(
                        explicit_inputs, explicit_outputs, explicit_input_offsets, explicit_output_offsets)
                    new_input = {**new_input, **explicit_index}
            # Make sure mapping is valid
            elif any(self._char_match_pattern.finditer(input_string)) or any(self._char_match_pattern.finditer(output_string)):
                raise MalformedMapping()
            # else just use default many-to-many indexing
            else:
                default_index = self.strings_to_lists(
                    input_string, output_string, input_index, output_index)
                new_input = {**new_input, **default_index}
            return {**intermediate_index, **new_input}

    def apply_rules(self, to_convert: str, index: bool = False, debugger: bool = False) -> Union[str, Tuple[str, Indices]]:
        """ Apply all the rules in self.mapping sequentially.

        @param to_convert: str
            This is the string to convert

        @param index: bool
            This is whether to preserve indices, default is False

        @param debugger: bool
            This is whether to show intermediary steps, default is False

        """
        indices = {}
        rules_applied = []

        if not self.case_sensitive:
            to_convert = to_convert.lower()

        if self.norm_form:
            to_convert = normalize(to_convert, self.norm_form)

        # initialized converted
        converted = to_convert

        if index:
            input_index = 0
            output_index = 0
            new_index = {}
            for char in range(len(to_convert)):
                # account for many-to-many rules making the input index
                # outpace the char-by-char conversion
                if char < input_index:
                    continue
                if not char in new_index or new_index[char]['input_string'] != to_convert[char]:
                    input_index = char
                    new_index[char] = {'input_string': to_convert[char],
                                       'output': {}}
                # intermediate form refreshes on each new char
                intermediate_conversion = to_convert
                rule_applied = False
                # go through rules
                for io in self.mapping:
                    io_copy = copy.deepcopy(io)
                    # find all matches.
                    for match in io_copy['match_pattern'].finditer(intermediate_conversion):
                        match_index = match.start()
                        # if start index of match is equal to input index,
                        # then apply the rule and append the index-formatted tuple
                        # to the main indices list
                        if match_index == input_index:
                            if self.out_delimiter:
                                # Don't add the delimiter to the last segment
                                if not char + (len(io_copy['in']) - 1) >= len(to_convert) - 1:
                                    io_copy['out'] += self.out_delimiter
                            # convert the final output
                            output_sub = re.sub(
                                re.compile(r'{\d+}'), '', io_copy['out'])
                            intermediate_output = intermediate_conversion[:char] + re.sub(
                                io_copy["match_pattern"], output_sub, intermediate_conversion[char:])
                            if debugger and intermediate_conversion != intermediate_output:
                                applied_rule = {"input": intermediate_conversion,
                                                "rule": io_copy, "output": intermediate_output}
                                rules_applied.append(applied_rule)
                            # update intermediate converted form
                            intermediate_conversion = intermediate_output
                            # get the new index tuple
                            non_null_index = self.return_index(
                                input_index, output_index, io_copy['in'], io_copy['out'],
                                to_convert, new_index)
                            # if it's not empty, then a rule has applied and it can overwrite
                            # the previous intermediate index tuple
                            if non_null_index:
                                rule_applied = True
                                new_index = {**new_index, **non_null_index}
                        # if you've gone past the input_index, you can safely break from the loop
                        elif match_index > input_index:
                            break
                # increase the index counters
                # new_index = self.convert_index_to_tuples(new_index)
                # if the rule applied
                if rule_applied and new_index[char]['output']:
                    # add the new index to the list of indices
                    indices = {**indices, **new_index}
                    # get the length of the new index inputs and outputs
                    # and increase the input counter by the length of the input
                    input_index = max(new_index.keys())
                    input_index += 1
                    # do the same with outputs
                    outputs = {}
                    for v in new_index.values():
                        outputs = {**outputs, **v['output']}
                    output_index = max(outputs.keys())
                    output_index += 1
                else:
                    # if a rule wasn't applied, just add on the input character
                    # as the next input and output character
                    new_index = {**new_index, **{input_index: {'input_string': to_convert[input_index],
                                                               'output': {output_index: to_convert[input_index]}}}}
                    # merge it
                    indices = {**indices, **new_index}
                    # add one to input and output
                    input_index += 1
                    output_index += 1
        else:
            # if not worrying about indices, just do the conversion rule-by-rule
            for io in self.mapping:
                io_copy = copy.deepcopy(io)
                if self.out_delimiter:
                    io_copy['out'] += self.out_delimiter
                output_sub = re.sub(re.compile(r'{\d+}'), '', io_copy['out'])
                if re.search(io_copy["match_pattern"], converted):
                    inp = converted
                    outp = re.sub(
                        io_copy["match_pattern"], output_sub, converted)
                    if debugger and inp != outp:
                        applied_rule = {"input": inp,
                                        "rule": io_copy, "output": outp}
                        rules_applied.append(applied_rule)
                    converted = outp
            # Don't add the delimiter to the last segment
            converted = converted.rstrip()
        if index and debugger:
            io_states = Indices(indices)
            return (io_states.output(), io_states, rules_applied)
        if debugger:
            return (converted, rules_applied)
        if index:
            io_states = Indices(indices)
            return (io_states.output(), io_states)
        return converted


class CompositeTransducer():
    ''' Class containing one or more Transducers

    Attributes
    ----------

    transducers: List[Transducer]
        Ordered list of Transducer objects to concatenate.

    '''

    def __init__(self, transducers: List[Transducer]):
        self._transducers = transducers

    def __repr__(self):
        return f"{__class__} between {self._transducers[0].mapping.kwargs['in_lang']} and {self._transducers[-1].mapping.kwargs['out_lang']}"

    def __call__(self, to_convert: str, index: bool = False, debugger: bool = False):
        return self.apply_rules(to_convert, index, debugger)

    def apply_rules(self, to_convert: str, index: bool = False, debugger: bool = False):
        converted = to_convert
        indexed = []
        debugged = []
        for transducer in self._transducers:
            response = transducer(converted, index, debugger)
            if index:
                indexed.append(response[1])
                if debugger:
                    debugged += response[2]
            if debugger:
                debugged += response[1]
            if index or debugger:
                converted = response[0]
            else:
                converted = response
        if index and debugger:
            return (converted, IndexSequence(*indexed), debugged)
        if index:
            return (converted, IndexSequence(*indexed))
        if debugger:
            return (converted, debugged)
        return converted



from unittest import main, TestCase
from g2p.mappings import Mapping
from g2p.transducer import CompositeTransducer, Transducer


class TransducerTest(TestCase):
    ''' Basic Transducer Test
    '''
    @classmethod
    def setUpClass(cls):
        cls.test_mapping_moh = Mapping(in_lang="moh", out_lang='moh-ipa')
        cls.test_mapping = Mapping([{'in': 'a', "out": 'b'}])
        cls.test_mapping_rev = Mapping(
            [{"in": 'a', "out": 'b'}], reverse=True)
        cls.test_mapping_ordered_feed = Mapping(
            [{"in": "a", "out": "b"}, {"in": "b", "out": "c"}], as_is=True)
        cls.test_mapping_ordered_counter_feed = Mapping(
            [{"in": "b", "out": "c"}, {"in": "a", "out": "b"}], as_is=True)
        cls.test_as_is_mapping = Mapping(
            [{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}], as_is=True)
        cls.test_not_as_is_mapping = Mapping(
            [{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}])
        cls.test_case_sensitive_mapping = Mapping(
            [{"in": "'n", "out": "n̓"}], case_sensitive=True)
        cls.test_case_insensitive_mapping = Mapping(
            [{"in": "'n", "out": "n̓"}], case_sensitive=False)
        cls.test_case_sensitive_transducer = Transducer(
            cls.test_case_sensitive_mapping)
        cls.test_case_insensitive_transducer = Transducer(
            cls.test_case_insensitive_mapping)
        cls.test_trans_as_is = Transducer(cls.test_as_is_mapping)
        cls.test_trans_not_as_is = Transducer(cls.test_not_as_is_mapping)
        cls.test_trans = Transducer(cls.test_mapping)
        cls.test_trans_ordered_feed = Transducer(
            cls.test_mapping_ordered_feed)
        cls.test_trans_ordered_counter_feed = Transducer(
            cls.test_mapping_ordered_counter_feed)
        cls.test_trans_rev = Transducer(cls.test_mapping_rev)
        cls.test_trans_moh = Transducer(cls.test_mapping_moh)
        cls.test_trans_composite = CompositeTransducer(
            [cls.test_trans, cls.test_trans_rev])
        cls.test_trans_composite_2 = CompositeTransducer(
            [cls.test_trans_rev, cls.test_trans])

    def test_ordered(self):
        transducer_i_feed = self.test_trans_ordered_feed('a', True)
        transducer_feed = self.test_trans_ordered_feed('a')
        transducer_i_counter_feed = self.test_trans_ordered_counter_feed(
            'a', True)
        transducer_counter_feed = self.test_trans_ordered_counter_feed('a')
        # These should feed b -> c
        self.assertEqual(transducer_feed, 'c')
        self.assertEqual(transducer_i_feed[1](), [((0, "a"), (0, "c"))])
        # These should counter-feed b -> c
        self.assertEqual(transducer_counter_feed, 'b')
        self.assertEqual(transducer_i_counter_feed[1](), [
                         ((0, "a"), (0, "b"))])

    def test_forward(self):
        self.assertEqual(self.test_trans('a'), "b")
        self.assertEqual(self.test_trans('b'), "b")

    def test_backward(self):
        self.assertEqual(self.test_trans_rev("b"), 'a')
        self.assertEqual(self.test_trans_rev("a"), 'a')

    def test_lang_import(self):
        self.assertEqual(self.test_trans_moh('kawennón:nis'), 'kawẽnõːnis')

    def test_composite(self):
        self.assertEqual(self.test_trans_composite('aba'), 'aaa')
        self.assertEqual(self.test_trans_composite_2('aba'), 'bbb')

    def test_as_is(self):
        self.assertEqual(self.test_trans_as_is("'y"), "jˀ")
        self.assertEqual(self.test_trans_not_as_is("'y"), "ʣˀ")

    def test_case_sensitive(self):
        self.assertEqual(self.test_case_sensitive_transducer("'N"), "'N")
        self.assertEqual(self.test_case_sensitive_transducer("'n"), "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'N"), "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'n"), "n̓")


if __name__ == "__main__":
    main()



from unittest import main, TestCase

import requests

from g2p import APP
from g2p.log import LOGGER
from g2p.cli import update

class CliTest(TestCase):
    def setUp(self):
        self.runner = APP.test_cli_runner()

    def test_update(self):
        result = self.runner.invoke(update)
        self.assertEqual(result.exit_code, 0)

if __name__ == '__main__':
    main()


import os
import click
import yaml
from flask.cli import FlaskGroup
from collections import OrderedDict
from networkx import draw

from g2p.mappings.create_fallback_mapping import align_to_dummy_fallback
from g2p.mappings.langs import cache_langs, LANGS_NETWORK
from g2p.mappings.create_ipa_mapping import create_mapping
from g2p.mappings.utils import is_ipa, is_xsampa
from g2p.mappings import Mapping
from g2p._version import VERSION
from g2p import APP, SOCKETIO, make_g2p


def create_app():
    return APP


@click.version_option(version=VERSION, prog_name="g2p")
@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    '''Management script for G2P'''

@click.option('--ipa/--no-ipa', default=False)
@click.option('--dummy/--no-dummy', default=False)
@click.argument('in_lang', type=click.Choice([x for x in LANGS_NETWORK.nodes if not is_ipa(x) and not is_xsampa(x)]))
@cli.command()
def generate_mapping(in_lang, dummy, ipa):
    ''' Generate English mapping
    '''
    if not ipa and not dummy:
        click.echo('You have to choose to generate either an IPA-based mapping or a dummy fallback mapping. Check the docs for more information.')
    if ipa:
        eng_ipa = Mapping(in_lang='eng-ipa', out_lang='eng-arpabet')
        new_mapping = Mapping(in_lang=in_lang, out_lang=f'{in_lang}-ipa')
        click.echo(f"Writing English IPA mapping for {in_lang} to file")
        create_mapping(new_mapping, eng_ipa, write_to_file=True)
    if dummy:
        new_mapping = Mapping(in_lang=in_lang, out_lang=f'{in_lang}-ipa')
        click.echo(f"Writing dummy fallback mapping for {in_lang} to file")
        dummy_config, dummy_mapping = align_to_dummy_fallback(new_mapping, write_to_file=True)

@click.argument('path', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@cli.command()
def generate_mapping_network(path):
    ''' Generate a png of the network of mapping languages. Requires matplotlib.
    '''
    import matplotlib.pyplot as plt
    draw(LANGS_NETWORK, with_labels=True)
    plt.show()


@click.argument('out_lang', type=click.Choice(LANGS_NETWORK.nodes))
@click.argument('in_lang', type=click.Choice(LANGS_NETWORK.nodes))
@click.argument('input_text', type=click.STRING)
@cli.command()
def convert(in_lang, out_lang, input_text):
    ''' Convert any text
    '''
    transducer = make_g2p(in_lang, out_lang)
    click.echo(transducer(input_text))


@cli.command()
def update():
    ''' Update cached language files
    '''
    cache_langs()
